/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationAuthority;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.*;

/**
 * Verwaltet und dokumentiert die Zustnde von Tasks und deren Fortschritt. Bei Zustandswechseln einzelner Tasks und in zyklischen Abstnden werden die Zustnde
 * aller Tasks ausgegeben und ber den Datenverteiler publiziert. Der Fortschritt von Aktionen einzelner Tasks wird zyklisch ausgegeben.
 *
 * @author Kappich Systemberatung
 * @version $Revision: none $
 */
public class ArchiveLongTermTaskStatePublisher implements LongTermTaskStatePublisher {

	/** Intervalldauer zur Ausgabe und Publikation der Zustnde aller Tasks */
	private static final long ALL_TASK_UPDATE_INTERVAL = 60000;

	/** Intervalldauer zur Ausgabe des Zustands einer Task */
	private static final long TASK_UPDATE_INTERVAL = 10000;

	
	
	private List<ArchiveLongTermTask> _tasks = new ArrayList<ArchiveLongTermTask>();

	private ClientDavInterface _davConnection = null;

	private long _nextAllTaskUpdateTime;

	private boolean _enableDataSource = false;

	private Debug _debug = Debug.getLogger();

	private static final String NEWLINE = System.getProperty("line.separator");

	private ArchiveManager _archMgr;

	private final String _minorChangeHeaderText;

	private final String _majorChangeHeaderText;

	private AttributeGroup _atg;

	private ConfigurationAuthority _authority;

	private DataDescription _dataDescription;

	public ArchiveLongTermTaskStatePublisher(final ArchiveManager archMgr, final String minorChangeHeaderText, final String majorChangeHeaderText) {
		_archMgr = archMgr;
		_minorChangeHeaderText = minorChangeHeaderText;
		_majorChangeHeaderText = majorChangeHeaderText;
		_nextAllTaskUpdateTime = System.currentTimeMillis();
	}

	public void setDavConnection(final ClientDavInterface davConnection) {
		_davConnection = davConnection;
		try {
			final DataModel dataModel = davConnection.getDataModel();
			_atg = dataModel.getAttributeGroup("atg.hintergrundTaskZustnde");
			final Aspect asp = dataModel.getAspect("asp.archivSystem");
			if(_atg == null || asp == null)
			{
				_debug.info("Kann Taskzustnde nicht verffentlichen, bentigte Attributgruppe und/oder Aspekt sind nicht im Datenmodell vorhanden. Es wird mindestens Version 23 von kb.systemModellGlobal.config bentigt.");
				return;
			}
			_authority = _archMgr.getConfigAuth();
			_dataDescription = new DataDescription(_atg, asp);
			davConnection.subscribeSender(new ClientSender(), _authority, _dataDescription, SenderRole.source());
		}
		catch(OneSubscriptionPerSendData oneSubscriptionPerSendData) {
			_debug.warning("Es gibt bereits ein Objekt, dass Taskzustnde verffentlicht", oneSubscriptionPerSendData);
		}
		catch(Exception e){
			_debug.warning("Fehler beim Anmelden der Verffentlichung der Taskzustnde", e);
		}
	}

	private Data createDataObject() {
		final Data data = _davConnection.createData(_atg);
		final Data.Array array = data.getItem("HintergrundTaskZustnde").asArray();
		int i = 0;
		array.setLength(_tasks.size());
		for(ArchiveLongTermTask task : _tasks) {
			insertData(task, array.getItem(i++));
		}
		return data;
	}

	private void insertData(final ArchiveLongTermTask task, final Data item) {
		synchronized(task){
			item.getItem("TaskName").asTextValue().setText(task._taskName);
			item.getItem("TaskZustand").asTextValue().setText(task._taskState.getStateName());
			item.getItem("TaskFortschrittsBeschreibung").asTextValue().setText(task.getProgress());
			item.getItem("TaskFortschritt").asUnscaledValue().set(task._progress) ;
			item.getItem("TaskFortschrittMaximum").asUnscaledValue().set(task._maximumProgress);
			item.getItem("TaskFortschrittErgebnis").asUnscaledValue().set(task._progressResult);
		}
	}


	public Task registerTask(final String taskName) {
		synchronized(this) {
			ArchiveLongTermTask longTermTask = new ArchiveLongTermTask(taskName);
			_tasks.add(longTermTask);
			return longTermTask;
		}
	}

	private class ClientSender implements ClientSenderInterface{
		public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
			synchronized(ArchiveLongTermTaskStatePublisher.this) {
				if(object.equals(_authority) && dataDescription.equals(_dataDescription)) {
					_enableDataSource = (state == START_SENDING);
					if(_enableDataSource) {
						try {
							final Data data = createDataObject();
							_davConnection.sendData(new ResultData(_authority, _dataDescription, System.currentTimeMillis(), data));
							_debug.info("Taskzustnde werden ber Datenverteiler verffentlicht");
						}
						catch(Exception e) {
							_debug.warning("Fehler beim Senden der Taskzustnde", e);
						}
					}
					else {
						_debug.info("Verffentlichung der Taskzustnde gestoppt");
					}
				}
			}
		}

		public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
			return true;
		}
	}

	private class ArchiveLongTermTask implements LongTermTaskStatePublisher.Task {

		private long _nextTaskUpdateTime;

		private final String _taskName;

		private TaskState _taskState = TaskState.INIT;

		private boolean _minorChanges = false;

		private String _format = "";

		private long _progress = -1;

		private long _maximumProgress = -1;

		private long _progressResult = -1;

		public ArchiveLongTermTask(final String taskName) {
			_taskName = taskName;
			_nextTaskUpdateTime = System.currentTimeMillis() + TASK_UPDATE_INTERVAL;
		}

		public void setTaskState(final TaskState taskState) {
			synchronized(ArchiveLongTermTaskStatePublisher.this) {
				synchronized(this) {
					if(_taskState != taskState) {
						final long now = System.currentTimeMillis();
						if(_minorChanges) publishMinorChange(now);
						_taskState = taskState;
						_format = "";
						_progress = -1;
						_maximumProgress = -1;
						_progressResult = -1;
						_minorChanges = false;
						publishMajorChanges(now);
					}
				}
			}
		}

		public void setTaskProgress(final String format, final long progress, final long maximumProgress, final long progressResult) {
			// Die Werte werden ohne Prfung auf nderung bernommen, weil sie ohnehin nur in vorgegebenen Zeitabstnden ausgegeben werden
			final long now = System.currentTimeMillis();
			boolean publish = false;
			synchronized(this) {
				_format = format;
				_progress = progress;
				_maximumProgress = maximumProgress;
				_progressResult = progressResult;
				_minorChanges = true;
				if(now > _nextTaskUpdateTime) {
					publish = true;
				}
			}
			if(publish) {
				synchronized(ArchiveLongTermTaskStatePublisher.this) {
					notifyMinorChanges(now);
				}
				synchronized(this) {
					if(now > _nextTaskUpdateTime) {
						publishMinorChange(now);
					}
				}
			}
		}

		private void publishMinorChange(final long now) {
			_debug.info(this.toString());
			publishedMinorChanges(now);
		}

		@Override
		public String toString() {
			final StringBuilder stringBuilder = new StringBuilder();
			stringBuilder.append(_minorChangeHeaderText);
			stringBuilder.append(NEWLINE);
			appendTo(stringBuilder);
			return stringBuilder.toString();
		}

		public void appendTo(final StringBuilder stringBuilder) {
			stringBuilder.append(_taskName);
			stringBuilder.append(": ");
			stringBuilder.append(_taskState.getStateName());
			if(!_format.equals("")) {
				stringBuilder.append(": ");
				final Formatter formatter = new Formatter(stringBuilder);
				formatter.format(_format, _progress, _maximumProgress, _progressResult);
			}
		}

		public String getProgress(){
			final Formatter formatter = new Formatter();
			formatter.format(_format, _progress, _maximumProgress, _progressResult);
			return formatter.toString();
		}

		public void publishedMinorChanges(final long now) {
			_nextTaskUpdateTime = now + TASK_UPDATE_INTERVAL;
			_minorChanges = false;
		}
	}

	private void publishMajorChanges(final long now) {
		final StringBuilder stringBuilder = new StringBuilder();
		stringBuilder.append(_majorChangeHeaderText);
		stringBuilder.append(NEWLINE);
		for(ArchiveLongTermTask longTermTask : _tasks) {
			synchronized(longTermTask) {
				longTermTask.appendTo(stringBuilder);
				stringBuilder.append(NEWLINE);
				longTermTask.publishedMinorChanges(now);
			}
		}
		_debug.info(stringBuilder.toString());

		if(_enableDataSource)
		{
			try {
				final Data data = createDataObject();
				_davConnection.sendData(new ResultData(_authority, _dataDescription, System.currentTimeMillis(), data));
			}
			catch(Exception e){
				_debug.warning("Fehler beim Senden der Taskzustnde", e);
			}
		}

		_nextAllTaskUpdateTime = now + ALL_TASK_UPDATE_INTERVAL;
	}

	private void notifyMinorChanges(final long now) {
		if(now > _nextAllTaskUpdateTime) {
			publishMajorChanges(now);
		}
	}
}
