/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.commands;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTreeIterator;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveQueryTask;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveTask;
import de.bsvrz.ars.ars.mgmt.tasks.LZVTask;
import de.bsvrz.ars.ars.mgmt.tasks.scheduler.TaskScheduler;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.ContainerSettings;
import de.bsvrz.sys.funclib.losb.datk.ContainerSettings.ContSettingsExc;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.CmdInterpreter;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.CmdMenu;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.Command;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.HiddenCommand;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.InfoCommands;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.net.UnknownHostException;
import java.text.DecimalFormat;
import java.util.*;


/**
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 6419 $ / $Date: 2009-03-10 23:15:57 +0100 (Di, 10 Mrz 2009) $ / ($Author: rs $)
 */
public class ArSCmdInterface {

	/** Max. Zeit der Inaktivitt, nach der der Server die Verbindung zum Client beendet. */
	public static int SEND_TIMEOUT = 60 * 5;

	/** Das Beenden-Kommando ist auf der ersten Menueebene unter dieser Nummer zu finden: */
	private static int QUIT_CMD_NR = 5;

	private ArchiveManager archMgr;

	private CmdInterpreter cmdInt;

	private int cmdPort, timeout;

	/**
	 * Erzeugt den Telnet Server. Startet ihn jedoch nicht - dazu muss {@link #start()} aufgerufen werden.
	 *
	 * @param am   ArchivManager
	 * @param port Port des Telnet Servers
	 */
	public ArSCmdInterface(ArchiveManager am, int port) {
		timeout = 10;
		cmdPort = port;
		archMgr = am;
		cmdInt = new CmdInterpreter(cmdPort, timeout);
		CmdMenu root = new CmdMenu("Archivsystem", "");

		//Submen Einstellungen
		CmdMenu subMenu = new CmdMenu("Einstellungen", "");
		root.addNode(subMenu);
		CmdMenu debugLevelMenu = InfoCommands.getDebugLevelMenu();
		subMenu.addNode(debugLevelMenu);
		subMenu.addCmd(
				new Command("Debug-Level ausgeben", "") {
					public void execute() throws Exception {
						println("Angemeldete Logger:\n" + Debug.debugInfo());
					}
				}
		);

		// Archivinformationen
		subMenu = new CmdMenu("Archivinformationen", "");
		root.addNode(subMenu);

		// Details einer Datenidentifikation abrufen (manuelle Navigation)
		subMenu.addCmd(new DataIdentInfo(archMgr));

		subMenu.addCmd(
				new Command("Auslastung der Archivierungs-Warteschlangen", "Beenden mit 'q'") {
					public void execute() throws Exception {
						println("Ausgabe kann mit 'q' beendet werden.");
						println("Die Ausgabe wird unterdrckt, wenn sie sich zur vorherigen nicht gendert hat.");
						println("Aktualisierungsrate in Millisekunden:");
						ArchiveMonitor am = null;
						try {
							int n = Integer.parseInt(readln());
							am = new ArchiveMonitor(archMgr, n, this);
							am.start();
						}
						catch(Exception e) {
							println("Ungltige Eingabe!");
						}
						while(true) {
							if(readln().equalsIgnoreCase("q")) {
								if(am != null) am.terminate();
								break;
							}
						}
					}
				}
		);

		subMenu.addCmd(InfoCommands.memoryInfo());
		subMenu.addCmd(InfoCommands.runtimeInfo());
		subMenu.addCmd(InfoCommands.systemInfo());

		subMenu.addCmd(
				new Command("Archiveinstellungen", "") {
					public void execute() throws Exception {
						TaskScheduler ts = archMgr.getTaskScheduler();
						printlnPlain(
								"TMax:                    " + Util.relTimestrMillis(archMgr.getInQueuesMgr().getArchiveDataReceiver().getTMaxATimeForwardStep())
						);
						printlnPlain("Restkapazitaet:          " + Util.kiloBlocks(LZVTask.getMinRemainingDiskCapa()) + " Bytes");
						printlnPlain("Sicherungstask");
						printlnPlain("  Letzter Lauf:          " + ts.getBackupLastRun());
						printlnPlain("  Aktuelle Einstellung:  " + ts.getBackupIntervals());
						printlnPlain("  Naechster Lauf:        " + ts.getBackupNextRun());
						printlnPlain("");
						printlnPlain("Loeschtask");
						printlnPlain("  Letzter Lauf:          " + ts.getDeleteLastRun());
						printlnPlain("  Aktuelle Einstellung:  " + ts.getDeleteIntervals());
						printlnPlain("  Naechster Lauf:        " + ts.getDeleteNextRun());
						printlnPlain("");
						printlnPlain("Nachforderungstask");
						printlnPlain("  Letzter Lauf:          " + ts.getRequestLastRun());
						printlnPlain("  Aktuelle Einstellung:  " + ts.getRequestIntervals());
						printlnPlain("  Naechster Lauf:        " + ts.getRequestNextRun());
						printlnPlain("");
						printlnPlain("Loeschzeitverlaengerung");
						printlnPlain("  Letzter Lauf:          " + ts.getLZVLastRun());
						printlnPlain("  Aktuelle Einstellung:  " + ts.getLZVIntervals());
						printlnPlain("");
						printlnPlain("Anzahl Bytes:            " + ArchiveQueryTask.getBufferSizeMux());
						printlnPlain("Anzahl Blocks:           " + ArchiveQueryTask.getBlockingFactorMux());
						printlnPlain("");
						printlnPlain("par. Anfr. hohe Prio.:   " + archMgr.getInQueuesMgr().getHiQueryTaskNum());
						printlnPlain("par. Anfr. mitt. Prio.:  " + archMgr.getInQueuesMgr().getMidQueryTaskNum());
						printlnPlain("par. Anfr. niedr. Prio.: " + archMgr.getInQueuesMgr().getLoQueryTaskNum());
					}
				}
		);

		subMenu.addCmd(
				new Command("Container Abschlussparameter", "") {
					public void execute() throws Exception {
						ContainerSettings settings = ArchiveTask.getContainerSettings();
						printlnPlain("Anz. Datensaetze = " + Util.kiloBlocks(settings.stdCloseConditions.maxContAnzDS) + " Stk.");
						printlnPlain("Cont. Groesse    = " + Util.kiloBlocks(settings.stdCloseConditions.maxContSize) + " Byte");
						printlnPlain("Zeitspanne Cont. = " + Util.relTimestr(settings.stdCloseConditions.maxContTime));
						if(settings.exceptions.size() > 0) {
							printlnPlain(" Ausnahmen:");
							for(ContSettingsExc cSettingsEx : settings.exceptions) {
								for(Iterator it = cSettingsEx.atgSet.iterator(); it.hasNext();) {
									Long appId = (Long)it.next();
									if(it.hasNext()) {
										printlnPlain("  " + archMgr.getAtg(appId).getPidOrNameOrId() + ",");
									}
									else {
										printlnPlain("  " + archMgr.getAtg(appId).getPidOrNameOrId() + " = ");
										printlnPlain(
												"      " + Util.kiloBlocks(cSettingsEx.excCloseConditions.maxContAnzDS) + " Stk., "
												+ Util.kiloBlocks(cSettingsEx.excCloseConditions.maxContSize) + " Byte, "
												+ Util.relTimestr(cSettingsEx.excCloseConditions.maxContTime)
										);
									}
								}
							}
						}
					}
				}
		);

		subMenu.addCmd(
				new Command("Informationen zum DataIdentTree", "") {
					public void execute() throws Exception {
						DataIdentTreeIterator iter = archMgr.getDataIdentTree().createIterator();
						long stDin, stParam;
						stDin = stParam = 0;
						long[] stats = new long[4]; // 3=Anz,Eintraege,Groesse  4=OA,ON,NA,NN
						double[] avgRatio = new double[4];
						double[] maxRatio = new double[4];
						ArchiveDataKind[] adks = Util.getAllDataKinds();
						String[] adk = new String[]{"OA", "ON", "NA", "NN"};
						DecimalFormat df = new DecimalFormat("0");

						while(iter.getNext()) {
							DataIdentNode din = iter.getDataIdentNode();
							stDin++;		  // Anzahl DIN
							if(din.isArSParameterized()) stParam++;    // Anzahl DIN isParameterized()

							for(int i = 0; i < 4; i++) {
								if(din.indexesAssigned(adks[i])) {
									stats[i]++;
									double fillRatio = din.getIndexFillRatio(adks[i]);
									avgRatio[i] += fillRatio;
									maxRatio[i] = Math.max(maxRatio[i], fillRatio);
								}
							}
						}
						printlnPlain("Datenidentifikationen");
						printlnPlain("=====================");
						printlnPlain("  Anzahl DId im Speicher: " + Util.leadBlank(stDin, 7) + "  gesamt");
						printlnPlain("                          " + Util.leadBlank(stParam, 7) + "  parametriert");
						for(int i = 0; i < adk.length; i++) {
							printlnPlain(
									"  Anzahl Indexe " + adk[i] + ":       " + Util.leadBlank(stats[i], 7) + "  avg: "
									+ (stats[i] != 0 ? Util.leadBlank(df.format(100 * avgRatio[i] / stats[i]), 3) + "%" : "-") + "  max: " + (stats[i] != 0 ?
									                                                                                                          Util.leadBlank(
											                                                                                                          df.format(
													                                                                                                          100
													                                                                                                          * maxRatio[i]
											                                                                                                          ), 3
									                                                                                                          ) + "%" : "-")
							);
						}
					}
				}
		);

		subMenu = new CmdMenu("System Info", "");
		root.addNode(subMenu);

		subMenu.addCmd(
				new Command("Command Line Arguments", "") {
					public void execute() throws Exception {
						List<String> ocla = archMgr.getOriginalCommandLineArguments();
						for(String arg : ocla) {
							printlnPlain("  " + arg);
						}
					}
				}
		);

		subMenu.addCmd(
				new Command("Environment", "") {
					public void execute() throws Exception {
						Map map = System.getenv();
						for(Iterator it = map.entrySet().iterator(); it.hasNext();) {
							Map.Entry entry = (Map.Entry)it.next();
							printlnPlain("  " + entry.getKey() + "=" + entry.getValue());
						}
					}
				}
		);

		subMenu.addCmd(
				new Command("System Properties", "") {
					public void execute() throws Exception {
						Properties props = System.getProperties();
						for(Enumeration enu = props.keys(); enu.hasMoreElements();) {
							String key = (String)enu.nextElement();
							printlnPlain("  " + key + "=" + props.getProperty(key));
						}
					}
				}
		);

		root.addCmd(
				new Command("Converter ID<->PID", "") {
					public void execute() throws Exception {
						printlnPlain("Bitte geben Sie eine ID oder eine PID ein.");
						String eingabe = readln();
						if(eingabe == null || eingabe.trim().length() == 0) {
							printlnPlain("Bitte geben Sie einen Wert ein.");
						}
						else {
							eingabe = eingabe.trim();
							try { // Erst gucken ob's eine Zahl ist -> ID
								int id = Integer.parseInt(eingabe);
								SystemObject obj = archMgr.getObj(id);
								if(obj != null) {
									printlnPlain("  ID: " + id + "   PID: " + obj.getPid() + "   Name: " + obj.getName());
								}
								else {
									printlnPlain("  Die ID " + eingabe + " ist nicht bekannt.");
								}
							}
							catch(Exception e) {  // keine Zahl
								try { // jetzt als text versuchen -> PID
									SystemObject obj = archMgr.getObj(eingabe);
									if(obj != null) {
										printlnPlain("  PID: " + eingabe + "   ID: " + obj.getId() + "   Name: " + obj.getName());
									}
									else {
										printlnPlain("  Die PID " + eingabe + " ist nicht bekannt.");
									}
								}
								catch(Exception ex) {
									printlnPlain("Unbekannter Fehler: " + ex.getMessage());
								}
							}
						}
					}
				}
		);

		root.addCmd(
				new Command("Archivsystem stoppen", "") {
					public void execute() throws Exception {
						println("Soll das Archivsystem beendet werden? [j|n]");
						if(readln().equals("j")) {
							Debug.getLogger().info("Archvisystem per Kommandoverbindung beendet");
							archMgr.quit();
						}
					}
				}
		);

		//"Geheime" Kommandos
		root.addCmd(			//Kommando hinzufgen, das nicht angezeigt wird
		                        new HiddenCommand("Gesamtanzahl der online-aktuellen Datenstze", 131008) {
			                        public void execute() throws Exception {
				                        int sum = 0;
				                        float[][] res = archMgr.getInQueuesMgr().getArchiveQueuesOnlineStatus();

				                        for(int i = 0; i < res.length; i++) {
					                        sum += res[i][0];
				                        }
				                        println("" + sum);
			                        }
		                        }
		);
		cmdInt.setMenu(root);
	}

	/** Startet den Telnet Server. */
	public void start() {
		cmdInt.start();
	}

	/**
	 * Beendet das Archivsystem ber den Telnet Server.
	 *
	 * @param host Host des Servers
	 * @param port Port des Servers
	 *
	 * @throws UnknownHostException Unbekannter Host.
	 * @throws IOException          Fehler beim erzeugen des Sockets.
	 */
	public static void sendQuitCmd(String host, int port) throws UnknownHostException, IOException {
		Socket connection = new Socket(host, port);
		connection.setSoTimeout(SEND_TIMEOUT * 1000);
		BufferedWriter out = new BufferedWriter(new OutputStreamWriter(connection.getOutputStream()));
		BufferedReader in = new BufferedReader(new InputStreamReader(connection.getInputStream()));

		readPrompt(in);
		write(out, QUIT_CMD_NR + "");
		readPrompt(in);
		write(out, "j");
		in.readLine();
	}

	private static void readPrompt(BufferedReader in) throws IOException {
		while(!in.readLine().endsWith(CmdInterpreter.PROMPT)) ;
	}

	private static void write(BufferedWriter out, String s) throws IOException {
		out.write(s);
		out.newLine();
		out.flush();
	}
}
