/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.sys.funclib.debug.Debug;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

/**
 * Ein ContainerBackupIndex ist eine Klasse, die innerhalb einer Datenindentifikationsverzeichnisses speichert, welche Container in diesem Verzeichnis noch zu
 * sichern sind, damit nicht jeder einzelne ContainerHEader geprft werden muss.<br> Es wird eine Datei namens _backupindex.idx gespeichert, welche die
 * eigentlichen Eintrge enthlt. Diese besteht aus einem Header-Bereich un einem Datenbereich, wobei der Datenbereich unterteilt ist, in sortierte Daten und
 * unsortierte Daten. Zur Optimierung des Zugriffs, wird, falls es notwendig wird, der komplette Datenbereich sortiert.<br> Als Daten enthlt ein
 * ContainerBackupIndex ContainerIDs mit einem zugehrigen Byte-Wert, der den Status dieses Container-Eintrags angibt (sichern oder ungltig). Sichern bedeutet,
 * dass der Container gesichert wird. Ungltig bedeutet, dass ein Eintrag ungltig ist. Ein Eintrag wird typischerweise als ungltig markiert, wenn der
 * entsprechende Container gesichert wurde, um diesen nicht erneut zu sichern. Einen Eintrag als ungltig zu markieren ist performancemig sinnvoller, als
 * stndig den gesamten Datenbereich neu zu schreiben um eventuelle Lcken zu schlieen.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public final class ContainerBackupIndex {

	/** Maximale MedienID unbekannt */
	public static final int MEDIUM_ID_UNKNOWN = -1;

	/** Datenidentifikation wurde noch nie gesichert */
	public static final int MEDIUM_ID_UNSAVED = -2;

	/** Dateiname unter dem der Index gespchert wird */
	public static final String BACKUP_INDEX_FILE_NAME = "_backupindex.idx";

	private static final Debug _debug = Debug.getLogger();

	private final ContainerBackupIndexImplementation _impl;

	private final Object _implLock = new Object();

	private boolean _isOpened = true;

	/**
	 * Gibt einen ContainerBackupIndex aus einem angegebenen Verzeichnis zurck
	 *
	 * @param path Verzeichnis
	 *
	 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
	 */
	public static ContainerBackupIndex getContainerBackupIndex(File path) {
		final ContainerBackupIndexImplementation implementation = ContainerBackupIndexImplementation.getContainerBackupIndex(path);
		if(implementation == null) return null;
		return new ContainerBackupIndex(implementation);
	}

	/**
	 * Gibt einen ContainerBackupIndex aus einem angegebenen Verzeichnis zurck
	 *
	 * @param path Verzeichnis
	 *
	 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
	 *
	 * @throws IOException IO-Fehler
	 */
	public static ContainerBackupIndex getContainerBackupIndex(String path) throws IOException {
		final ContainerBackupIndexImplementation implementation = ContainerBackupIndexImplementation.getContainerBackupIndex(path);
		if(implementation == null) return null;
		return new ContainerBackupIndex(implementation);
	}

	/**
	 * Erstellt einen neuen, leeren ContainerBackupIndex im angegebenen Verzeichnis. Ein eventuell bereits vorhandener wird berschrieben.
	 *
	 * @param path Verzeichnis
	 *
	 * @return der neue ContainerBackupIndex
	 *
	 * @throws IOException IO-Fehler
	 */
	public static ContainerBackupIndex createNewContainerBackupIndex(File path) throws IOException {
		return new ContainerBackupIndex(ContainerBackupIndexImplementation.createNewContainerBackupIndex(path));
	}

	/**
	 * Lscht den ContainerBackupIndex im angebenden Verzeichnis
	 *
	 * @param path Verzeichnis
	 *
	 * @return true wenn das Lschen erfolgreich war und der Index vorher existierte
	 */
	public static boolean deleteContainerBackupIndex(File path) {
		return ContainerBackupIndexImplementation.deleteContainerBackupIndex(path);
	}

	/**
	 * Prft ob im angebenden Verzeichnis ein ContainerBackupIndex existiert
	 *
	 * @param path Verzeichnis
	 *
	 * @return true wenn er existiert
	 */
	public static boolean existsIn(File path) {
		return ContainerBackupIndexImplementation.getContainerBackupIndex(path) != null;
	}

	/**
	 * Prft ob im angebenden Verzeichnis ein ContainerBackupIndex existiert
	 *
	 * @param path Verzeichnis
	 *
	 * @return true wenn er existiert
	 */
	public static boolean existsIn(String path) {
		return ContainerBackupIndexImplementation.getContainerBackupIndex(path) != null;
	}

	/**
	 * Erstellt mit der angebenden ContainerBackupIndexImplementation einen neuen ContainerBackupIndex.
	 *
	 * @param implementation Implementation
	 */
	private ContainerBackupIndex(ContainerBackupIndexImplementation implementation) {
		_impl = implementation;
	}

	/**
	 * Gibt alle zu sichernden ContainerIDs im Index zurck
	 *
	 * @return Liste mit ContainerIDs
	 *
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public List<Long> getContainerIDs() throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.getContainerIDs();
		}
	}

	/**
	 * Entfernt eine ContainerID aus dem Index
	 *
	 * @param containerID ContainerID die zu entfernen ist.
	 *
	 * @return True falls erfolgreich, false falls nicht gefunden
	 *
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public boolean removeContainerID(long containerID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.removeContainerID(containerID, MEDIUM_ID_UNKNOWN);
		}
	}

	/**
	 * Entfernt eine ContainerID aus dem Index
	 *
	 * @param containerID ContainerID die zu entfernen ist.
	 * @param mediumID    Beim sichern eines Containers kann eine MediumID mit angegeben werden, die dann im Index vermerkt wird. Wird nur bercksichtigt, wenn es
	 *                    sich um eine positive Zahl handelt und nicht bereits eine grere MedienID in der Datei vorhanden ist
	 *
	 * @return True falls erfolgreich, false falls nicht gefunden
	 *
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public boolean removeContainerID(long containerID, final int mediumID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.removeContainerID(containerID, mediumID);
		}
	}

	/**
	 * Fgt dem Index eine ContainerID hinzu
	 *
	 * @param containerID ContainerID die einzufgen ist
	 *
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public void appendContainerID(long containerID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			_impl.appendContainerID(containerID);
		}
	}

	/**
	 * Gibt die in der ContainerBackupIndex-Datei maximal gespeicherte MedienID zurck.
	 *
	 * @return MedienID
	 *
	 * @throws IOException IO-Fehler
	 */
	public int getMaxMedienID() throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.getMaxMedienID();
		}
	}

	/**
	 * Setzt die MedienID auf einen festgelegten Wert (erlaubt auch das Zurcksetzen oder das Setzen von {@link ContainerBackupIndex#MEDIUM_ID_UNSAVED} bzw. {@link
	 * ContainerBackupIndex#MEDIUM_ID_UNKNOWN}
	 *
	 * @param mediumID MedienID
	 *
	 * @throws IOException IO-Fehler
	 */
	public void setMaxMedienID(int mediumID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			_impl.setMaxMedienID(mediumID);
		}
	}

	/** Schliet diesen ContainerBackupIndex. Muss immer nach Benutzung aufgerufen werden um kein mgliches Speicherleck zu verursachen */
	public void close() {
		synchronized(_implLock) {
			if(_isOpened) {
				_impl.removeReference();
				_isOpened = false;
			}
		}
	}

	@Override
	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}

	@Override
	public String toString() {
		synchronized(_implLock) {
			return "ContainerBackupIndex{" + _impl + ", open=" + _isOpened + '}';
		}
	}

	/**
	 * Eigentliche Implementierung des ContainerBackupIndex. wird von der {@link ContainerBackupIndex}-Klasse gekapselt um einen sicheren Referenzzhler zu
	 * ermglichen. Zu den Zwecken der Threadsicherheit wird fr jede ContainerBackupIndex-Datei maximal eine Instanz erzeugt und verwaltet. Jede Instanz kann von
	 * mehreren ContainerBackupIndex-Objekten verwendet werden. Zu diesem Zweck implementiert diese Klasse einen Referenzzhler.
	 */
	private static final class ContainerBackupIndexImplementation {

		private static final int CURRENT_FILE_VERSION = 1;

		private File _file;

		private final Object _fileLock = new Object();

		/** Legt fest, dass der Container im Index geschlossen ist und gesichert werden kann (z.B. wenn die Parametrierung es verlangt) */
		private static final int CONTAINER_TO_SAVE = 0;

		/**
		 * Markiert einen ungltigen Index-Eintrag. Wenn ein Container aus dem Index gelscht wird, wird er zunchst mit diesem Flag markiert Erst wenn die Anzahl der
		 * ungltigen Eintrge einen bestimmten Wert berschreitet, wird die gesamte Datei neu geschrieben/komprimiert.
		 */
		private static final int CONTAINER_DELETED_FROM_INDEX = -1;

		/**
		 * Aus Synchronittsgrnden muss sichergestellt sein, dass wenn eine Indexdatei an zwei verschiedenen Stellen gleichzeitig angefragt wird, beide Threads das
		 * selbe Objekt erhalten. Daher werden hier die offenen Indexdateien verwaltet.
		 */
		private static final HashMap<File, ContainerBackupIndexImplementation> _openFiles = new HashMap<File, ContainerBackupIndexImplementation>();

		/** Anzahl der Referenzen */
		private int _referenceCount;

		private static final int SIZE_INTEGER = 4;

		private static final int SIZE_BYTE = 1;

		private static final int SIZE_LONG = 8;

		private static final int HEADER_VERSION_SIZE = SIZE_INTEGER;

		private static final int HEADER_MEDIUM_ID_SIZE = SIZE_INTEGER;

		private static final int HEADER_NUM_SORTED_ENTRIES_SIZE = SIZE_INTEGER;

		private static final int HEADER_NUM_GAPS_SIZE = SIZE_INTEGER;

		private static final int HEADER_POS_MEDIUM_ID = HEADER_VERSION_SIZE;

		private static final int HEADER_POS_NUM_SORTED_ENTRIES = HEADER_POS_MEDIUM_ID + HEADER_MEDIUM_ID_SIZE;

		private static final int HEADER_POS_NUM_GAPS = HEADER_POS_NUM_SORTED_ENTRIES + HEADER_NUM_SORTED_ENTRIES_SIZE;

		private static final int HEADER_SIZE = HEADER_VERSION_SIZE + HEADER_MEDIUM_ID_SIZE + HEADER_NUM_SORTED_ENTRIES_SIZE + HEADER_NUM_GAPS_SIZE;

		private static final int ENTRY_SIZE = SIZE_LONG + SIZE_BYTE;

		private ContainerBackupIndexImplementation(final File file) {
			_file = file;
			_referenceCount = 1;
			synchronized(_openFiles) {
				_openFiles.put(_file, this);
			}
		}

		private void addReference() {
			synchronized(_openFiles) {
				_referenceCount++;
			}
		}

		private void removeReference() {
			synchronized(_openFiles) {
				_referenceCount--;
				if(_referenceCount == 0) {
					_openFiles.remove(_file);
					_file = null;
				}
			}
		}

		/**
		 * Gibt eine ContainerBackupIndexImplementation aus einem angegebenen Verzeichnis zurck
		 *
		 * @param path Verzeichnis
		 *
		 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
		 */
		public static ContainerBackupIndexImplementation getContainerBackupIndex(final File path) {
			final File file = new File(path, BACKUP_INDEX_FILE_NAME);
			if(file.exists()) {
				return getInstance(file);
			}
			return null;
		}

		/**
		 * Gibt eine ContainerBackupIndexImplementation aus einem angegebenen Verzeichnis zurck
		 *
		 * @param path Verzeichnis
		 *
		 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
		 */
		public static ContainerBackupIndexImplementation getContainerBackupIndex(final String path) {
			final File file = new File(path, BACKUP_INDEX_FILE_NAME);
			if(file.exists()) {
				return getInstance(file);
			}
			return null;
		}

		/**
		 * Holt zu einer angegebenen Datei eine ContainerBackupIndexImplementation-Instanz. Entweder aus der _openFiles-HashMap oder es wird ein neues Objekt
		 * erstellt
		 *
		 * @param file Datei
		 *
		 * @return die ContainerBackupIndexImplementation
		 */
		private static ContainerBackupIndexImplementation getInstance(File file) {
			synchronized(_openFiles) {
				final ContainerBackupIndexImplementation instance = _openFiles.get(file);
				if(instance != null) {
					instance.addReference();
					return instance;
				}
				return new ContainerBackupIndexImplementation(file);
			}
		}

		/**
		 * Erstellt eine neuen, leeren ContainerBackupIndexImplementation im angegebenen Verzeichnis. Ein eventuell bereits vorhandener wird berschrieben.
		 *
		 * @param path Verzeichnis
		 *
		 * @return der neue ContainerBackupIndex
		 *
		 * @throws IOException IO-Fehler
		 */
		public static ContainerBackupIndexImplementation createNewContainerBackupIndex(final File path) throws IOException {
			synchronized(_openFiles) {
				final File file = new File(path, BACKUP_INDEX_FILE_NAME);
				ContainerBackupIndexImplementation containerBackupIndex = getContainerBackupIndex(file);
				if(containerBackupIndex == null) containerBackupIndex = new ContainerBackupIndexImplementation(file);
				containerBackupIndex.reset();
				return containerBackupIndex;
			}
		}

		/**
		 * Lscht einen ContainerBackupIndex aus dem angegeben Verzeichnis
		 *
		 * @param path Verzeichnis
		 *
		 * @return true wenn das Lschen erfolgreich war und die Datei vorher existierte
		 */
		public static boolean deleteContainerBackupIndex(final File path) {
			final File file = new File(path, BACKUP_INDEX_FILE_NAME);
			return file.delete();
		}

		/**
		 * Entfernt alle Daten aus der ContainerBackupIndexImplementation
		 *
		 * @throws IOException IO-Fehler
		 */
		private void reset() throws IOException {
			synchronized(_fileLock) {
				final RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw");
				try {
					randomAccessFile.setLength(0L);
					randomAccessFile.writeInt(CURRENT_FILE_VERSION);
					randomAccessFile.writeInt(MEDIUM_ID_UNKNOWN);
					randomAccessFile.writeInt(0);
					randomAccessFile.writeInt(0);
				}
				finally {
					randomAccessFile.close();
				}
			}
		}

		/**
		 * Fgt einen Container hinzu
		 *
		 * @param containerID Container-ID
		 *
		 * @throws IOException IO-Fehler
		 */
		public void appendContainerID(long containerID) throws IOException {
			synchronized(_fileLock) {
				final RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw");
				try {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					randomAccessFile.seek(HEADER_POS_NUM_SORTED_ENTRIES);
					final int numOfSortedEntries = randomAccessFile.readInt();
					if(randomAccessFile.length() == HEADER_SIZE + numOfSortedEntries * ENTRY_SIZE) {
						// Optimierung:
						// Es gibt keine unsortierten Eintrge in der Datei.
						// Falls der einzufgende Eintrag grer ist, als der letzte Eintrag, bleibt die komplette Containerliste sortiert und man kann
						// NUM_SORTED_ENTRIES um eins erhhen.
						final long lastContainerID;
						if(numOfSortedEntries == 0) {
							// Wenn es keine Eintrge gibt ist die Datei beim Einfgen des ersten immer sortiert
							lastContainerID = -1;
						}
						else {
							// Wenn es bereits Eintrge gibt muss der letzte untersucht werden
							randomAccessFile.seek(randomAccessFile.length() - ENTRY_SIZE);
							lastContainerID = randomAccessFile.readLong();
						}
						if(containerID > lastContainerID) {
							// Die neue ContainerID ist grer als die letzte, also NUM_SORTED_ENTRIES inkrementieren
							randomAccessFile.seek(HEADER_POS_NUM_SORTED_ENTRIES);
							randomAccessFile.writeInt(numOfSortedEntries + 1);
						}
					}
					randomAccessFile.seek(randomAccessFile.length());
					randomAccessFile.writeLong(containerID);
					randomAccessFile.writeByte(CONTAINER_TO_SAVE);
				}
				finally {
					randomAccessFile.close();
				}
			}
		}

		/**
		 * Entfernt eine ContainerID aus dem Index
		 *
		 * @param containerID ContainerID die zu entfernen ist.
		 * @param mediumID    MediumID im Header, die evtl. erhht werden muss
		 *
		 * @return True falls erfolgreich, false falls nicht gefunden
		 *
		 * @throws IOException IO-Fehler
		 */
		public boolean removeContainerID(long containerID, int mediumID) throws IOException {
			synchronized(_fileLock) {
				final RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw");
				try {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					final int fileMediumID = randomAccessFile.readInt();
					if(mediumID != MEDIUM_ID_UNSAVED && mediumID != MEDIUM_ID_UNKNOWN && mediumID > fileMediumID) {
						randomAccessFile.seek(HEADER_POS_MEDIUM_ID);
						randomAccessFile.writeInt(mediumID);
					}
					final int numOfSortedEntries = randomAccessFile.readInt();
					final int numOfGaps = randomAccessFile.readInt();
					final int numOfUnsortedEntries = (int)((randomAccessFile.length() - HEADER_SIZE) / ENTRY_SIZE - numOfSortedEntries);

					if((numOfUnsortedEntries > 500 && numOfUnsortedEntries * 3 > numOfSortedEntries) || (numOfGaps > 4000
					                                                                                     && numOfGaps * 3 > numOfSortedEntries * 2)) {
						// Wenn die Anzahl der deaktivierten Eintrge (Lcken) 4000 berschreitet und die Anzahl der Lcken mindestens zwei Drittel
						// der Eintrge ausmacht
						// sollte die Datei beim nchsten Zugriff "defragmentiert", also sortiert und komprimiert neu aufgebaut werden.
						// Das gleiche soll bei mehr als 500 unsortierten Eintrgen passieren, sofern diese mindestens ein drittel so gro sind, wie die Anzahl
						// der sortierten Eintrge.
						return removeContainerIDWithRebuild(containerID, randomAccessFile);
					}
					else {
						return removeContainerIDFast(containerID, randomAccessFile, numOfSortedEntries, numOfGaps);
					}
				}
				finally {
					randomAccessFile.close();
				}
			}
		}

		public int getMaxMedienID() throws IOException {
			synchronized(_fileLock) {
				final RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw");
				try {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					return randomAccessFile.readInt();
				}
				finally {
					randomAccessFile.close();
				}
			}
		}

		public void setMaxMedienID(int mediumID) throws IOException {
			synchronized(_fileLock) {
				final RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw");
				try {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					randomAccessFile.writeInt(mediumID);
				}
				finally {
					randomAccessFile.close();
				}
			}
		}

		/**
		 * Markiert einen Container-Eintrag als ungltig
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Eintrge
		 * @param numOfGaps          Anzahl ungltige Eintrge
		 *
		 * @return True wenn der Container gelscht wurde
		 *
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDFast(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries, final int numOfGaps) throws IOException {

			boolean entryRemoved = removeContainerIDInSortedArea(containerID, randomAccessFile, numOfSortedEntries, numOfGaps);

			if(!entryRemoved) {
				entryRemoved = removeContainerIDInUnsortedArea(containerID, randomAccessFile, numOfSortedEntries, numOfGaps);
			}
			return entryRemoved;
		}

		/**
		 * Lscht einen Container-Eintrag aus dem sortierten Bereich
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Eintrge
		 * @param numOfGaps          Anzahl ungltige Eintrge
		 *
		 * @return True wenn der Container gelscht wurde
		 *
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDInSortedArea(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries, final int numOfGaps) throws IOException {
			int min = 0;
			int max = numOfSortedEntries - SIZE_BYTE;
			int mid;
			while(max >= min) {

				mid = (max + min) / 2;
				randomAccessFile.seek(HEADER_SIZE + mid * ENTRY_SIZE); // zum Eintrag gehen
				final long idAtMid = randomAccessFile.readLong();

				if(containerID < idAtMid) {
					max = mid - SIZE_BYTE;
				}
				else if(containerID > idAtMid) {
					min = mid + SIZE_BYTE;
				}
				else {
					if(randomAccessFile.readByte() != CONTAINER_DELETED_FROM_INDEX) {
						randomAccessFile.seek(randomAccessFile.getFilePointer() - SIZE_BYTE);
						//gefunden
						randomAccessFile.writeByte(CONTAINER_DELETED_FROM_INDEX);
						randomAccessFile.seek(HEADER_POS_NUM_GAPS);
						randomAccessFile.writeInt(numOfGaps + 1);
						return true;
					}
					break;
				}
			}
			return false;
		}

		/**
		 * Lscht einen Container-Eintrag aus dem unsortierten Bereich
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Eintrge
		 * @param numOfGaps          Anzahl ungltige Eintrge
		 *
		 * @return True wenn der Container gelscht wurde
		 *
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDInUnsortedArea(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries, final int numOfGaps) throws IOException {
			randomAccessFile.seek(HEADER_SIZE + numOfSortedEntries * ENTRY_SIZE);
			while(randomAccessFile.getFilePointer() < randomAccessFile.length()) {
				final long id = randomAccessFile.readLong();
				if(randomAccessFile.readByte() != CONTAINER_DELETED_FROM_INDEX && id == containerID) {
					randomAccessFile.seek(randomAccessFile.getFilePointer() - SIZE_BYTE);
					randomAccessFile.writeByte(CONTAINER_DELETED_FROM_INDEX);
					randomAccessFile.seek(HEADER_POS_NUM_GAPS);
					randomAccessFile.writeInt(numOfGaps + 1);
					return true;
				}
			}
			return false;
		}

		/**
		 * Lscht einen Container-Eintrag aus der Datei und sortiert dabei den kompletten Datenbereich neu
		 *
		 * @param containerIdToRemove ContainerID
		 * @param randomAccessFile    geffnete Datei
		 *
		 * @return True wenn der Container gelscht wurde
		 *
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDWithRebuild(
				final long containerIdToRemove, final RandomAccessFile randomAccessFile) throws IOException {

			_debug.fine("Baue Index neu auf");

			final ArrayList<IndexEntry> entryList = new ArrayList<IndexEntry>((int)(_file.length() / ENTRY_SIZE));
			boolean entryRemoved = false;
			// Alle Eintrge einlesen...
			while(randomAccessFile.getFilePointer() < randomAccessFile.length()) {
				final long id = randomAccessFile.readLong();
				final byte type = randomAccessFile.readByte();
				if(type == CONTAINER_DELETED_FROM_INDEX) continue; // Gelschte Eintrge ignorieren
				if(id == containerIdToRemove) {
					// Den zu lschenden Container ebenso
					entryRemoved = true;
					continue;
				}
				entryList.add(new IndexEntry(id, type));
			}
			Collections.sort(entryList);
			// ...und jetzt die Datei neu schreiben
			randomAccessFile.setLength(HEADER_SIZE + entryList.size() * ENTRY_SIZE);
			randomAccessFile.seek(HEADER_POS_NUM_SORTED_ENTRIES);
			randomAccessFile.writeInt(entryList.size());
			randomAccessFile.writeInt(0);
			for(IndexEntry entry : entryList) {
				randomAccessFile.writeLong(entry.getContainerID());
				randomAccessFile.writeByte(entry.getType());
			}
			return entryRemoved;
		}

		@Override
		public String toString() {
			synchronized(_fileLock) {
				return _file.toString();
			}
		}

		/**
		 * Gibt alle ContainerIDs im Index zurck
		 *
		 * @return Liste mit ContainerIDs
		 *
		 * @throws IOException IO-Fehler
		 */
		public List<Long> getContainerIDs() throws IOException {
			synchronized(_fileLock) {
				final RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "r");
				try {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);					
					randomAccessFile.seek(HEADER_SIZE);
					final ArrayList<Long> fileList = new ArrayList<Long>((int)((randomAccessFile.length() - HEADER_SIZE) / ENTRY_SIZE));
					while(randomAccessFile.getFilePointer() < randomAccessFile.length()) {
						final long id = randomAccessFile.readLong();
						final byte type = randomAccessFile.readByte();
						if(type == CONTAINER_DELETED_FROM_INDEX) continue;
						fileList.add(id);
					}
					return fileList;
				}
				finally {
					randomAccessFile.close();
				}
			}
		}

		/** Hilfsklasse die einen einzelnen Dateneintrag speichert. Wird zum Sortieren benutzt. */
		private final static class IndexEntry implements Comparable {

			private long _containerID;

			private byte _type;

			public IndexEntry(long containerID, byte type) {
				this._containerID = containerID;
				this._type = type;
			}

			public int compareTo(Object o) {
				return ((Long)_containerID).compareTo(((IndexEntry)o).getContainerID());
			}

			public long getContainerID() {
				return _containerID;
			}

			public byte getType() {
				return _type;
			}

			@Override
			public String toString() {
				return "IndexEntry{" + "_containerID=" + _containerID + ", _type=" + _type + '}';
			}
		}
	}
}

