/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence.index;

import java.util.*;

/**
 * Ergebnis einer Indexanfrage. Enthlt eine Liste von Container-IDs jeweils mit Min/Max-Werten. Bei Anfragen an den monotonen Index sind Min/Max-Werte fuer
 * Datenindex, Archivzeit und Datenzeit enthalten, bei Anfragen an den Datenzeitindex nur Min/Max-Werte fuer Datenzeit und Datenindex (die Archivzeit wurde nicht
 * benoetigt und deshalb auch nicht im Datenzeitindex gespeichert). Im letzteren Fall sind die Min/Max-Werte der Archivzeit auf {@link #NO_VALUE} gesetzt.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 9776 $ / $Date: 2011-12-19 05:58:31 +0100 (Mo, 19 Dez 2011) $ / ($Author: rs $)
 */
public class IndexResult {

	public static final int NO_VALUE = -1;

	public static IndexResult EMPTY;

	static {
		EMPTY = new IndexResult();
		EMPTY.result = new ArrayList<long[]>(0);	// Speicher sparen...
	}

	private List<long[]> result = new ArrayList<long[]>(512);

	/**
	 * Fuegt einen Container mit den angegebenen Min/Max-Werten der Ergebnisliste hinzu.
	 *
	 * @param cID   Container-ID
	 * @param dIMin Datenindex-Minimum des Containers
	 * @param dIMax Datenindex-Maximum des Containers
	 * @param aTMin Archivzeit-Minimum des Containers
	 * @param aTMax Archivzeit-Maximum des Containers
	 * @param dTMin Datenzeit-Minimum des Containers
	 * @param dTMax Datenzeit-Maximum des Containers
	 */
	public void add(long cID, long dIMin, long dIMax, long aTMin, long aTMax, long dTMin, long dTMax) {
		result.add(new long[]{cID, dIMin, dIMax, aTMin, aTMax, dTMin, dTMax});
	}

	/**
	 * Fuegt einen Container mit den angegebenen Min/Max-Werten der Ergebnisliste hinzu.
	 *
	 * @param cID   Container-ID
	 * @param dTMin Datenzeit-Minimum des Containers
	 * @param dTMax Datenzeit-Maximum des Containers
	 * @param dIMin Datenindex-Minimum des Containers
	 * @param dIMax Datenindex-Maximum des Containers
	 */
	public void add(long cID, long dTMin, long dTMax, long dIMin, long dIMax) {
		result.add(new long[]{cID, dIMin, dIMax, NO_VALUE, NO_VALUE, dTMin, dTMax});
	}

	public void updateDTMax(int pos, long dtMax) {
		assert pos < result.size() : "Ungltige Position";
		long[] tmp = result.get(pos);
		tmp[6] = dtMax;
		result.set(pos, tmp);
	}


	public void updateDIMax(int pos, long diMax) {
		assert pos < result.size() : "Ungltige Position";
		long[] tmp = result.get(pos);
		tmp[2] = diMax;
		result.set(pos, tmp);
	}

	/**
	 * Fuegt einen Container mit den angegebenen Min/Max-Werten der Ergebnisliste hinzu.
	 *
	 * @param cID   Container-ID
	 * @param dIMin Datenindex-Minimum des Containers
	 * @param dIMax Datenindex-Maximum des Containers
	 */
	public void add(long cID, long dIMin, long dIMax) {
		result.add(new long[]{cID, dIMin, dIMax, NO_VALUE, NO_VALUE, NO_VALUE, NO_VALUE});
	}

	/**
	 * Fuegt einen Container mit den angegebenen Min/Max-Werten in der Ergebnisliste an angegebener Stelle hinzu
	 *
	 * @param pos
	 * @param cID   Container-ID
	 * @param dTMin Datenzeit-Minimum des Containers
	 * @param dTMax Datenzeit-Maximum des Containers
	 * @param dIMin Datenindex-Minimum des Containers
	 * @param dIMax Datenindex-Maximum des Containers
	 */
	public void insert(int pos, long cID, long dTMin, long dTMax, long dIMin, long dIMax) {
		result.add(pos, new long[]{cID, dIMin, dIMax, NO_VALUE, NO_VALUE, dTMin, dTMax});
	}

	/**
	 * Fuegt einen Container mit den angegebenen Min/Max-Werten in der Ergebnisliste an angegebener Stelle hinzu
	 *
	 * @param pos
	 * @param cID   Container-ID
	 * @param dIMin Datenindex-Minimum des Containers
	 * @param dIMax Datenindex-Maximum des Containers
	 */
	public void insert(int pos, long cID, long dIMin, long dIMax) {
		result.add(pos, new long[]{cID, dIMin, dIMax, NO_VALUE, NO_VALUE, NO_VALUE, NO_VALUE});
	}

	/** @return Anzahl der enthaltenen Containerinformationen */
	public int size() {
		return result.size();
	}

	/** @return Wahr, wenn keine Container in der Ergebnisliste enthalten sind, falsch sonst */
	public boolean isEmpty() {
		return size() == 0;
	}

	/**
	 * @param pos Position in der Ergebnisliste
	 *
	 * @return ContainerID des Eintrags an der angegebenen Position
	 */
	public long getCID(int pos) {
		return pos >= result.size() ? NO_VALUE : result.get(pos)[0];
	}

	/**
	 * @param pos Position in der Ergebnisliste
	 *
	 * @return Datenindex-Minimum des Eintrags an der angegebenen Position
	 */
	public long getDIMin(int pos) {
		return pos >= result.size() ? NO_VALUE : result.get(pos)[1];
	}

	/**
	 * @param pos Position in der Ergebnisliste
	 *
	 * @return Datenindex-Maximum des Eintrags an der angegebenen Position
	 */
	public long getDIMax(int pos) {
		return pos >= result.size() ? NO_VALUE : result.get(pos)[2];
	}

	/**
	 * @param pos Position in der Ergebnisliste
	 *
	 * @return Archivzeit-Minimum des Eintrags an der angegebenen Position
	 */
	public long getATMin(int pos) {
		return pos >= result.size() ? NO_VALUE : result.get(pos)[3];
	}

	/**
	 * @param pos Position in der Ergebnisliste
	 *
	 * @return Archivzeit-Maximum des Eintrags an der angegebenen Position
	 */
	public long getATMax(int pos) {
		return pos >= result.size() ? NO_VALUE : result.get(pos)[4];
	}

	/**
	 * @param pos Position in der Ergebnisliste
	 *
	 * @return Datenzeit-Minimum des Eintrags an der angegebenen Position
	 */
	public long getDTMin(int pos) {
		return pos >= result.size() ? NO_VALUE : result.get(pos)[5];
	}

	/**
	 * @param pos Position in der Ergebnisliste
	 *
	 * @return Datenzeit-Maximum des Eintrags an der angegebenen Position
	 */
	public long getDTMax(int pos) {
		return pos >= result.size() ? NO_VALUE : result.get(pos)[6];
	}

	/* (non-Javadoc)
		 * @see java.lang.Object#toString()
		 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for(long[] l : result) {
			sb.append(
					"{cID=" + l[0] + ", dIMin=" + l[1] + ", dIMax=" + l[2] + ", aTMin=" + l[3] + ", aTMax=" + l[4] + ", dTMin=" + l[5] + ", dTMax=" + l[6]
					+ "} "
			);
		}
		return sb.toString();
	}
}
